#pragma once

#include "debug.hpp"

#include <boost/function.hpp>

#include "config.hpp"
#include "dpt.hpp"
#include "group.hpp"

namespace knx
{

class connection_private;

/**
 * \brief This class handles the ip connection(s) to the knx gateway
 */
class KNX_IMPORT_EXPORT connection
{
public:
    /**
     * @brief creates the connection with given config
     */
    connection ( const knx::config & config );

    //! @brief simple destructor
    ~connection();

    /**
     * @brief blocking start of the connection background thread
     *
     * This function is a blocking call to start the background thread.
     * It is normally only called by the handler.
     *
     * @warning you should not call this function from your code.
     */
    void start();

    /**
     * @brief non-blocking stop of the background thread
     *
     * This function send the termination signals and waits for the thread to stop.
     */
    void stop();

    /**
     * @brief central function to set any knx group value on the bus
     *
     * This function is used to set any value on the knx bus.
     */
    template<typename data_type>
    void set ( knx::group group, typename data_type::major_type::set_type data );

    /**
     * @brief central function to get any knx group value from the bus
     *
     * This function is used to get any value from the knx bus.
     *
     * @param group A valid group id as string (e.g. "1/2/3")
     * @return true if read was successful -- false otherwise
     */
    template<typename data_type>
    bool get ( knx::group group, typename data_type::major_type::set_type & data );

    /**
     * @brief central function to continously get any knx group value from the bus
     *
     * This function is used to listen for any value from the knx bus.
     *
     * A minimal example would be:
     *
     * @code
     * #include "knx.hpp"
     * #include "debug.hpp"
     * void my_callback ( knx::dpt_1_001::set_type & data )
     * {
     *     std::cout << "listened to " << data << std::endl;
     * }
     * int main ( int argc, char ** argv )
     * {
     *     knx::config config ( argc, argv );
     *     knx::connection connection ( config );
     *     knx::handle handle ( connection );
     *     boost::this_thread::sleep ( boost::posix_time::seconds ( 2 ) );
     *     connection.listen<knx::dpt_1_001> ( "0/0/1", my_callback );
     *     boost::this_thread::sleep( boost::posix_time::seconds ( 20 ) );
     *     return 0;
     * }
     * @endcode
     *
     * @param group A valid group id as string (e.g. "1/2/3")
     * @param callback A function to be called whenever data has been received
     * @return true if registering was successful -- false otherwise
     */
    template<typename data_type>
    bool listen ( knx::group group, boost::function<void ( typename data_type::major_type::set_type & ) > callback );


private:
    //! hidden implementation
    connection_private * connection_private_pointer;
    //! prevent copy
    connection ( connection & rhs );
    //! prevent copy
    connection & operator= ( connection & rhs );
};

}

