#pragma once

#include "debug.hpp"

#include <stdint.h>
#include <sstream>
#include <boost/date_time.hpp>
#include <boost/mpl/vector.hpp>

#pragma pack(1)

namespace knx
{

#define SUPPORTED_DPTS \
 (1_001) \
 (2_001) \
 (3_007) \
 (4_001) \
 (5_001) \
 (6_001) \
 (7_001) \
 (8_001) \
 (9_001) \
 (10_001) \
 (11_001) \
 (12_001) \
 (13_001) \
 (14_000) \
 (15_000) \
 (16_000) \
 (17_001) \
 (18_001) \
 (19_001) \
 (20_011) \
 (21_001) \
 (26_001) \
 (219_001) \
 (232_600)

// TODO MISSING DPTS
// 022.100
// 023.001
// 024.001 (VAR STRING)
// 027.001
// 028.001 (VAR UTF8)
// 029.010
// 031.101
// 200.100
// 201.100
// 202.001
// 203.002
// 204.001
// 205.002
// 206.100
// 207.100
// 209.100
// 210.100
// 211.100
// 212.100
// 213.100
// 214.100
// 215.100
// 216.100
// 217.001
// 218.001
// 220.100
// 221.001
// 222.100
// 223.100
// 224.100
// 225.001
// 229.001
// 230.1000
// 231.001
// 234.001

#define OP(s, data, elem) BOOST_PP_CAT(DPT_, elem)

enum dpt_index
{
    BOOST_PP_SEQ_ENUM(BOOST_PP_SEQ_TRANSFORM(OP, _, SUPPORTED_DPTS))
};

#define OP2(s, data, elem) BOOST_PP_CAT(knx::dpt_, elem)

#define SEQ BOOST_PP_SEQ_TRANSFORM(OP2, _, SUPPORTED_DPTS)

// template<uint32_t SIZE>
// class data_store
// {
// private:
//     uint8_t raw_data[SIZE];
// };

class dpt
{
public:
    template<typename data_type>
    static std::string describe ( const typename data_type::major_type::set_type & data );
};

/**
 * @brief major class for all 1.XXX data point types
 */
class KNX_IMPORT_EXPORT dpt_1 // : public data_store<1>
{
public:
    typedef uint8_t data_type;
    typedef bool set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * \brief data point type 1.001 simple boolean value
 */
class KNX_IMPORT_EXPORT dpt_1_001
{
public:
    static const int INDEX = DPT_1_001;
    typedef dpt_1 major_type;
    static const major_type::set_type ON;
    static const major_type::set_type OFF;
    static std::string describe ( const major_type::set_type & data );
};

//! data point type 1.001 is better known as switch
typedef dpt_1_001 dpt_switch;

/**
 * @brief major class for all 2.XXX data point types
 */
class KNX_IMPORT_EXPORT dpt_2
{
public:
    typedef uint8_t data_type;
    typedef std::pair<bool,bool> set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * \brief data point type 2.001 control value
 */
class KNX_IMPORT_EXPORT dpt_2_001
{
public:
    static const int INDEX = DPT_2_001;
    typedef dpt_2 major_type;
    static const major_type::set_type NO_CONTROL;
    static const major_type::set_type CONTROL_VALUE_ZERO;
    static const major_type::set_type CONTROL_VALUE_ONE;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 3.XXX data point types
 */
class KNX_IMPORT_EXPORT dpt_3
{
public:
    typedef uint8_t data_type;
    typedef std::pair<bool,uint8_t> set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * \brief data point type 3.007 dimming control
 */
class KNX_IMPORT_EXPORT dpt_3_007
{
public:
    static const int INDEX = DPT_3_007;
    typedef dpt_3 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 4.XXX data point types
 */
class KNX_IMPORT_EXPORT dpt_4
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        data ( uint8_t character = 0 ) :static_part ( 0 ),character ( character ) {}
        uint8_t character;
    };
    typedef data data_type;
    typedef uint8_t set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * \brief data point type 4.001 char ascii
 */
class KNX_IMPORT_EXPORT dpt_4_001
{
public:
    static const int INDEX = DPT_4_001;
    typedef dpt_4 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 5.XXX data point types
 */
class dpt_5
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        data ( uint8_t unsigned_value = 0 );
        uint8_t unsigned_value;
    };
    typedef data data_type;
    typedef uint8_t set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * \brief data point type 5.001 unsigned 8 bit
 */
class dpt_5_001
{
public:
    static const int INDEX = DPT_5_001;
    typedef dpt_5 major_type;
    typedef TYPENAME major_type::set_type set_type;
    static std::string describe ( const major_type::set_type & data );
};

//! 5.001 is better known as scaling value
typedef dpt_5_001 dpt_scaling;

/**
 * @brief major class for all 6.XXX signed 8 bit value
 */
class dpt_6
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        data ( int8_t real_signed_value = 0 ) :static_part ( 0 ),real_signed_value ( real_signed_value ) {}
        int8_t real_signed_value;
    };
    typedef data data_type;
    typedef int8_t set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * \brief data point type 6.001 percent v8
 */
class dpt_6_001
{
public:
    static const int INDEX = DPT_6_001;
    typedef dpt_6 major_type;
    typedef TYPENAME major_type::set_type set_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 7.XXX unsigned signed 16 bit value
 */
class dpt_7
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        data ( ) : static_part ( 0 ), value ( 0 ) {}
        uint16_t value;
    };
    typedef data data_type;
    typedef uint16_t set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * \brief data point type 7.001 unsigned 16bit pulse counter
 */
class dpt_7_001
{
public:
    static const int INDEX = DPT_7_001;
    typedef dpt_7 major_type;
    typedef TYPENAME major_type::set_type set_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 8.XXX signed 16 bit value
 */
class dpt_8
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        data ( ) : static_part ( 0 ), value ( 0 ) {}
        int16_t value;
    };
    typedef data data_type;
    typedef int16_t set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * \brief data point type 8.001 16bit signed pulse counter
 */
class dpt_8_001
{
public:
    static const int INDEX = DPT_8_001;
    typedef dpt_8 major_type;
    typedef TYPENAME major_type::set_type set_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 9.XXX data point types
 */
class dpt_9
{
public:
    static const uint16_t INVALID_DATA = 0x7fff;
    static const uint16_t MAX = 0x7ffe;
    static const uint16_t MIN = 0xf800;
    static const float MAX_FLOAT =  670760.96f;
    static const float MIN_FLOAT = -671088.64f;
    class data
    {
    private:
        uint8_t static_part;
    public:
        data ( float float_value = 0.f );
        uint16_t value;
    };
    typedef data data_type;
    typedef float set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * \brief data point type 9.001 value temp
 */
class dpt_9_001
{
public:
    static const int INDEX = DPT_9_001;
    typedef dpt_9 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 10.XXX data point types
 */
class dpt_10
{
public:
    //! @brief internal class storing a time value
    class time
    {
        friend std::ostream & operator << ( std::ostream & os, const time & time );
    public:
        inline bool operator== ( const time & other )
        {
            return ( ( this->day == other.day ) && ( this->hour == other.hour ) && ( this->minutes == other.minutes ) && ( this->seconds == other.seconds ) );
        }
        //! @brief creates a current time value
        time();
        uint8_t day;
        uint8_t hour;
        uint8_t minutes;
        uint8_t seconds;
        static const uint8_t NODAY = 0;
        static const uint8_t MONDAY = 1;
        static const uint8_t TUESDAY = 2;
        static const uint8_t WEDNESDAY = 3;
        static const uint8_t THURSDAY = 4;
        static const uint8_t FRIDAY = 5;
        static const uint8_t SATURDAY = 6;
        static const uint8_t SUNDAY = 7;
    };
    typedef uint32_t data_type;
    typedef time set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

std::ostream & operator << ( std::ostream & os, const dpt_10::time & time );

/**
 * @brief class holding a time value
 */
class KNX_IMPORT_EXPORT dpt_10_001
{
public:
    static const int INDEX = DPT_10_001;
    typedef dpt_10 major_type;
    static std::string describe ( const major_type::set_type & data );
    static major_type::time now();
};

//! 10.001 is known as time of day value including hours. minutes, seconds and day of the week
typedef dpt_10_001 dpt_timeofday;

/**
 * @brief major class for all 11.XXX data point types
 */
class dpt_11
{
public:
    //! @brief internal class storing a date value
    class date
    {
    public:
        inline bool operator== ( const date & other )
        {
            return ( ( this->day == other.day ) && ( this->month == other.month ) && ( this->year == other.year ) );
        }
        //! @brief creates a current date value
        date();
        uint8_t day;
        uint8_t month;
        uint8_t year;
    };
    typedef uint32_t data_type;
    typedef date set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a date value
 */
class KNX_IMPORT_EXPORT dpt_11_001
{
public:
    static const int INDEX = DPT_11_001;
    typedef dpt_11 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 12.XXX data point types
 */
class dpt_12
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        data ( ) : static_part ( 0 ), value ( 0 ) {}
        uint32_t value;
    };
    typedef data data_type;
    typedef uint32_t set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a 32 bit unsigned value
 */
class KNX_IMPORT_EXPORT dpt_12_001
{
public:
    static const int INDEX = DPT_12_001;
    typedef dpt_12 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 13.XXX data point types
 */
class dpt_13
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        data ( ) : static_part ( 0 ), value ( 0 ) {}
        int32_t value;
    };
    typedef data data_type;
    typedef uint32_t set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a 32 bit signed value
 */
class KNX_IMPORT_EXPORT dpt_13_001
{
public:
    static const int INDEX = DPT_13_001;
    typedef dpt_13 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 14.XXX data point types
 */
class dpt_14
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        data ( ) : static_part ( 0 ), value ( 0 ) {}
        float value;
    };
    typedef data data_type;
    typedef uint32_t set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a 32 bit float value
 */
class KNX_IMPORT_EXPORT dpt_14_000
{
public:
    static const int INDEX = DPT_14_000;
    typedef dpt_14 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 15.XXX data point types
 */
class dpt_15
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        uint8_t d65,d43,d21;
        uint8_t epdc_index;
    };
    //! TODO add documentation
    class set_data
    {
    public:
        uint8_t d6,d5,d4,d3,d2,d1;
        bool e,p,d,c;
        uint8_t index;
    };
    typedef data data_type;
    typedef set_data set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a access data type
 */
class KNX_IMPORT_EXPORT dpt_15_000
{
public:
    static const int INDEX = DPT_15_000;
    typedef dpt_15 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 16.XXX data point types
 */
class dpt_16
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        char text[14];
    };
    typedef data data_type;
    typedef std::string set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a fixed length string of up to 14 chars in ASCII
 */
class KNX_IMPORT_EXPORT dpt_16_000
{
public:
    static const int INDEX = DPT_16_000;
    typedef dpt_16 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 17.XXX data point types
 */
class dpt_17
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        uint8_t value;
    };
    typedef data data_type;
    typedef uint8_t set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a scene number
 */
class KNX_IMPORT_EXPORT dpt_17_001
{
public:
    static const int INDEX = DPT_17_001;
    typedef dpt_17 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 18.XXX data point types
 */
class dpt_18
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        uint8_t value;
    };
    class set_data
    {
    public:
        //! false corresponds to activate and true to learn the scene number
        bool c;
        uint8_t scene_number;
    };
    typedef data data_type;
    typedef set_data set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a scene control value
 */
class KNX_IMPORT_EXPORT dpt_18_001
{
public:
    static const int INDEX = DPT_18_001;
    typedef dpt_18 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 19.XXX data point types
 */
class dpt_19
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        uint32_t date;
        uint32_t time;
    };
    //! TODO
    class set_data
    {
    public:
        uint8_t year, month, day_of_month, day_of_week, hour_of_day, minutes, seconds;
        bool f, wd, nwd, ny, nd, ndow, nt, suti, clq;
    };
    typedef data data_type;
    typedef set_data set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a date time value
 */
class KNX_IMPORT_EXPORT dpt_19_001
{
public:
    static const int INDEX = DPT_19_001;
    typedef dpt_19 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 20.XXX data point types
 */
class dpt_20
{
public:
    typedef uint16_t data_type;
    typedef uint8_t set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a error class system value
 */
class KNX_IMPORT_EXPORT dpt_20_011
{
public:
    enum {
        NO_FAULT,
        GENERAL_DEVICE_FAULT,
        COMMUNICATION_FAULT,
        CONFIGURATION_FAULT,
        HARDWARE_FAULT,
        SOFTWARE_FAULT
    };
    static const int INDEX = DPT_20_011;
    typedef dpt_20 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 21.XXX data point types
 */
class dpt_21
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        uint8_t value;
    };
    class set_data
    {
    public:
        bool out_of_service;
        bool fault;
        bool overridden;
        bool in_alarm;
        bool alarm_un_ack;
    };
    typedef data data_type;
    typedef set_data set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a status gen value
 */
class KNX_IMPORT_EXPORT dpt_21_001
{
public:
    static const int INDEX = DPT_21_001;
    typedef dpt_21 major_type;
    static std::string describe ( const major_type::set_type & data );
};

// TODO VARSTRING 24.001 -- BAD THING

/**
 * @brief major class for all 26.XXX data point types
 */
class dpt_26
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        uint8_t value;
    };
    class set_data
    {
    public:
        uint8_t scene_number;
        bool b;
    };
    typedef data data_type;
    typedef set_data set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a scene info value
 */
class KNX_IMPORT_EXPORT dpt_26_001
{
public:
    static const int INDEX = DPT_26_001;
    typedef dpt_26 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 219.XXX data point types
 */
class dpt_219
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        uint8_t log_number;
        uint8_t alarm_priority;
        uint8_t application_area;
        uint8_t error_class;
        uint8_t attributes;
        uint8_t alarm_status_attributes;
    };
    typedef data data_type;
    typedef data set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding a alarm info
 */
class KNX_IMPORT_EXPORT dpt_219_001
{
public:
    static const int INDEX = DPT_219_001;
    typedef dpt_219 major_type;
    static std::string describe ( const major_type::set_type & data );
};

/**
 * @brief major class for all 232.XXX data point types
 */
class dpt_232
{
public:
    class data
    {
    private:
        uint8_t static_part;
    public:
        uint8_t r;
        uint8_t g;
        uint8_t b;
    };
    typedef data data_type;
    typedef data set_type;
    static void set ( set_type value, data_type & data );
    static set_type get ( const data_type & data );
};

/**
 * @brief class holding rgb color value
 */
class KNX_IMPORT_EXPORT dpt_232_600
{
public:
    static const int INDEX = DPT_232_600;
    typedef dpt_232 major_type;
    static std::string describe ( const major_type::set_type & data );
};

#define OP3(s, data, elem) BOOST_PP_CAT(dpt_, elem)

typedef boost::mpl::vector<
BOOST_PP_SEQ_ENUM(BOOST_PP_SEQ_TRANSFORM(OP3, _, SUPPORTED_DPTS))
> dpt_types;

}

#pragma pack()




